# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.texts.remote.cts
   :synopsis: CtsTextMetadata and CapitainsCtsPassage implementation for dealing with CTS API Responses

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""

from __future__ import unicode_literals

from MyCapytain.common.metadata import Metadata
from MyCapytain.common.utils import xmlparser
from MyCapytain.common.constants import XPATH_NAMESPACES, Mimetypes, RDF_NAMESPACES
from MyCapytain.common.reference import URN, Reference
from MyCapytain.resources.collections import cts as CtsCollection
from MyCapytain.resources.prototypes import text as prototypes
from MyCapytain.resources.texts.base.tei import TEIResource
from MyCapytain.errors import MissingAttribute


class __SharedMethod__(prototypes.InteractiveTextualNode):
    """ Set of methods shared by CtsTextMetadata and CapitainsCtsPassage

    :param retriever: Retriever used to retrieve other data
    :type retriever: MyCapytain.retrievers.prototypes.CitableTextServiceRetriever
    """

    @property
    def depth(self):
        """ Depth of the current opbject

        :return: Int representation of the depth based on URN information
        :rtype: int
        """
        if self.urn.reference:
            return len(self.urn.reference)

    def __init__(self, retriever=None, *args, **kwargs):
        super(__SharedMethod__, self).__init__(*args, **kwargs)
        self.__retriever__ = retriever
        self.__first__ = False
        self.__last__ = False
        if retriever is None:
            raise MissingAttribute("Object has not retriever")

    @property
    def retriever(self):
        """ Retriever object used to query for more data

        :rtype: CitableTextServiceRetriever
        """
        return self.__retriever__

    def getValidReff(self, level=1, reference=None):
        """ Given a resource, CitableText will compute valid reffs

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: Int
        :param reference: CapitainsCtsPassage reference
        :type reference: Reference
        :rtype: list(str)
        :returns: List of levels
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        if level == -1:
            level = len(self.citation)

        xml = self.retriever.getValidReff(
            level=level,
            urn=urn
        )
        xml = xmlparser(xml)
        self.__parse_request__(xml.xpath("//ti:request", namespaces=XPATH_NAMESPACES)[0])

        return [ref.split(":")[-1] for ref in xml.xpath("//ti:reply//ti:urn/text()", namespaces=XPATH_NAMESPACES)]

    def getTextualNode(self, subreference=None):
        """ Retrieve a passage and store it in the object

        :param subreference: Reference of the passage (Note : if given a list, this should be a list of string that \
        compose the reference)
        :type subreference: Union[Reference, URN, str, list]
        :rtype: CtsPassage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """
        if isinstance(subreference, URN):
            urn = str(subreference)
        elif isinstance(subreference, Reference):
            urn = "{0}:{1}".format(self.urn, str(subreference))
        elif isinstance(subreference, str):
            if ":" in subreference:
                urn = subreference
            else:
                urn = "{0}:{1}".format(self.urn.upTo(URN.NO_PASSAGE), subreference)
        elif isinstance(subreference, list):
            urn = "{0}:{1}".format(self.urn, ".".join(subreference))
        else:
            urn = str(self.urn)

        response = xmlparser(self.retriever.getPassage(urn=urn))

        self.__parse_request__(response.xpath("//ti:request", namespaces=XPATH_NAMESPACES)[0])
        return CtsPassage(urn=urn, resource=response, retriever=self.retriever)

    def getReffs(self, level=1, subreference=None):
        """ Reference available at a given level

        :param level: Depth required. If not set, should retrieve first encountered level (1 based)
        :type level: Int
        :param subreference: Subreference (optional)
        :type subreference: Reference
        :rtype: [text_type]
        :returns: List of levels
        """
        if self.depth is not None:
            level += self.depth

        return self.getValidReff(level, subreference)

    def getPassagePlus(self, reference=None):
        """ Retrieve a passage and informations around it and store it in the object

        :param reference: Reference of the passage
        :type reference: Reference or List of text_type
        :rtype: CtsPassage
        :returns: Object representing the passage
        :raises: *TypeError* when reference is not a list or a Reference
        """
        if reference:
            urn = "{0}:{1}".format(self.urn, reference)
        else:
            urn = str(self.urn)

        response = xmlparser(self.retriever.getPassagePlus(urn=urn))

        passage = CtsPassage(urn=urn, resource=response, retriever=self.retriever)
        passage.__parse_request__(response.xpath("//ti:reply/ti:label", namespaces=XPATH_NAMESPACES)[0])
        self.citation = passage.citation
        return passage

    def __parse_request__(self, xml):
        """ Parse a request with metadata information

        :param xml: LXML Object
        :type xml: Union[lxml.etree._Element]
        """
        for node in xml.xpath(".//ti:groupname", namespaces=XPATH_NAMESPACES):
            lang = node.get("xml:lang") or CtsText.DEFAULT_LANG
            self.metadata.add(RDF_NAMESPACES.CTS.groupname, lang=lang, value=node.text)
            self.set_creator(node.text, lang)

        for node in xml.xpath(".//ti:title", namespaces=XPATH_NAMESPACES):
            lang = node.get("xml:lang") or CtsText.DEFAULT_LANG
            self.metadata.add(RDF_NAMESPACES.CTS.title, lang=lang, value=node.text)
            self.set_title(node.text, lang)

        for node in xml.xpath(".//ti:label", namespaces=XPATH_NAMESPACES):
            lang = node.get("xml:lang") or CtsText.DEFAULT_LANG
            self.metadata.add(RDF_NAMESPACES.CTS.label, lang=lang, value=node.text)
            self.set_subject(node.text, lang)

        for node in xml.xpath(".//ti:description", namespaces=XPATH_NAMESPACES):
            lang = node.get("xml:lang") or CtsText.DEFAULT_LANG
            self.metadata.add(RDF_NAMESPACES.CTS.description, lang=lang, value=node.text)
            self.set_description(node.text, lang)

        # Need to code that p
        if self.citation.isEmpty() and xml.xpath("//ti:citation", namespaces=XPATH_NAMESPACES):
            self.citation = CtsCollection.XmlCtsCitation.ingest(
                xml,
                xpath=".//ti:citation[not(ancestor::ti:citation)]"
            )

    def getLabel(self):
        """ Retrieve metadata about the text

        :rtype: Metadata
        :returns: Dictionary with label informations
        """
        response = xmlparser(
            self.retriever.getLabel(urn=str(self.urn))
        )

        self.__parse_request__(
            response.xpath("//ti:reply/ti:label", namespaces=XPATH_NAMESPACES)[0]
        )

        return self.metadata

    def getPrevNextUrn(self, reference):
        """ Get the previous URN of a reference of the text

        :param reference: Reference from which to find siblings
        :type reference: Union[Reference, str]
        :return: (Previous CapitainsCtsPassage Reference,Next CapitainsCtsPassage Reference)
        """
        _prev, _next = __SharedMethod__.prevnext(
            self.retriever.getPrevNextUrn(
                urn="{}:{}".format(
                    str(
                        URN(
                            str(self.urn)).upTo(URN.NO_PASSAGE)
                    ),
                    str(reference)
                )
            )
        )
        return _prev, _next

    def getFirstUrn(self, reference=None):
        """ Get the first children URN for a given resource

        :param reference: Reference from which to find child (If None, find first reference)
        :type reference: Reference, str
        :return: Children URN
        :rtype: URN
        """
        if reference is not None:
            if ":" in reference:
                urn = reference
            else:
                urn = "{}:{}".format(
                    str(URN(str(self.urn)).upTo(URN.NO_PASSAGE)),
                    str(reference)
                )
        else:
            urn = str(self.urn)
        _first = __SharedMethod__.firstUrn(
            self.retriever.getFirstUrn(
                urn
            )
        )
        return _first

    @property
    def firstId(self):
        """ Children passage

        :rtype: str
        :returns: First children of the graph. Shortcut to self.graph.children[0]
        """
        if self.__first__ is False:
            # Request the next urn
            self.__first__ = self.getFirstUrn()
        return self.__first__

    @property
    def lastId(self):
        """ Children passage

        :rtype: str
        :returns: First children of the graph. Shortcut to self.graph.children[0]
        """
        if self.__last__ is False:
            # Request the next urn
            self.__last__ = self.childIds[-1]
        return self.__last__

    @staticmethod
    def firstUrn(resource):
        """ Parse a resource to get the first URN

        :param resource: XML Resource
        :type resource: etree._Element
        :return: Tuple representing previous and next urn
        :rtype: str
        """
        resource = xmlparser(resource)
        urn = resource.xpath("//ti:reply/ti:urn/text()", namespaces=XPATH_NAMESPACES, magic_string=True)

        if len(urn) > 0:
            urn = str(urn[0])
            return urn.split(":")[-1]

    @staticmethod
    def prevnext(resource):
        """ Parse a resource to get the prev and next urn

        :param resource: XML Resource
        :type resource: etree._Element
        :return: Tuple representing previous and next urn
        :rtype: (str, str)
        """
        _prev, _next = False, False
        resource = xmlparser(resource)
        prevnext = resource.xpath("//ti:prevnext", namespaces=XPATH_NAMESPACES)

        if len(prevnext) > 0:
            _next, _prev = None, None
            prevnext = prevnext[0]
            _next_xpath = prevnext.xpath("ti:next/ti:urn/text()", namespaces=XPATH_NAMESPACES, smart_strings=False)
            _prev_xpath = prevnext.xpath("ti:prev/ti:urn/text()", namespaces=XPATH_NAMESPACES, smart_strings=False)

            if len(_next_xpath):
                _next = _next_xpath[0].split(":")[-1]

            if len(_prev_xpath):
                _prev = _prev_xpath[0].split(":")[-1]

        return _prev, _next


class CtsText(__SharedMethod__, prototypes.CitableText):
    """ API CtsTextMetadata object

    :param urn: A URN identifier
    :type urn: Union[URN, str, unicode]
    :param resource: An API endpoint
    :type resource: CitableTextServiceRetriever
    :param citation: XmlCtsCitation for children level
    :type citation: XmlCtsCitation
    :param id: Identifier of the subreference without URN informations
    :type id: List

    """

    DEFAULT_LANG = "eng"

    def __init__(self, urn, retriever, citation=None, **kwargs):
        super(CtsText, self).__init__(retriever=retriever, urn=urn, citation=citation, **kwargs)

    @property
    def reffs(self):
        """ Get all valid reffs for every part of the CitableText

        :rtype: MyCapytain.resources.texts.tei.XmlCtsCitation
        """
        if self.citation.isEmpty():
            self.getLabel()
        return [
            reff for reffs in [self.getValidReff(level=i) for i in range(1, len(self.citation) + 1)] for reff in reffs
        ]

    @property
    def nextId(self):
        raise NotImplementedError

    @property
    def next(self):
        raise NotImplementedError

    @property
    def prev(self):
        raise NotImplementedError

    @property
    def prevId(self):
        raise NotImplementedError

    @property
    def siblingsId(self):
        raise NotImplementedError

    def export(self, output=Mimetypes.PLAINTEXT, exclude=None, **kwargs):
        """ Export the collection item in the Mimetype required.

        ..note:: If current implementation does not have special mimetypes, reuses default_export method

        :param output: Mimetype to export to (Uses Mimetypes)
        :type output: str
        :param exclude: Informations to exclude. Specific to implementations
        :type exclude: [str]
        :return: Object using a different representation
        """
        return self.getTextualNode().export(output, exclude)


class CtsPassage(__SharedMethod__, prototypes.Passage, TEIResource):
    """ CapitainsCtsPassage representing

    :param urn:
    :param resource:
    :param retriever:
    :param args:
    :param kwargs:
    """

    def __init__(self, urn, resource, *args, **kwargs):
        SuperKwargs = {key: value for key, value in kwargs.items() if key not in ["parent"]}
        super(CtsPassage, self).__init__(resource=resource, *args, **SuperKwargs)
        self.urn = urn

        # Could be set during parsing
        self.__nextId__ = False
        self.__prev__ = False
        self.__first__ = False
        self.__last__ = False

        self.__parse__()

    @property
    def id(self):
        return str(self.urn.reference)

    @property
    def prevId(self):
        """ Previous passage Identifier

        :rtype: CtsPassage
        :returns: Previous passage at same level
        """
        if self.__prev__ is False:
            # Request the next urn
            self.__prev__, self.__nextId__ = self.getPrevNextUrn(reference=self.urn.reference)
        return self.__prev__

    @property
    def parentId(self):
        """ Shortcut for getting the parent passage identifier

        :rtype: Reference
        :returns: Following passage reference
        """
        return str(self.urn.reference.parent)

    @property
    def nextId(self):
        """ Shortcut for getting the following passage identifier

        :rtype: Reference
        :returns: Following passage reference
        """
        if self.__nextId__ is False:
            # Request the next urn
            self.__prev__, self.__nextId__ = self.getPrevNextUrn(reference=self.urn.reference)
        return self.__nextId__

    @property
    def siblingsId(self):
        """ Shortcut for getting the previous and next passage identifier

        :rtype: Reference
        :returns: Following passage reference
        """
        if self.__nextId__ is False or self.__prev__ is False:
            self.__prev__, self.__nextId__ = self.getPrevNextUrn(reference=self.urn.reference)
        return self.__prev__, self.__nextId__

    def __parse__(self):
        """ Given self.resource, split information from the CTS API

        :return: None
        """
        self.response = self.resource
        self.resource = self.resource.xpath("//ti:passage/tei:TEI", namespaces=XPATH_NAMESPACES)[0]

        self.__prev__, self.__nextId__ = __SharedMethod__.prevnext(self.response)

        if self.citation.isEmpty() and len(self.resource.xpath("//ti:citation", namespaces=XPATH_NAMESPACES)):
            self.citation = CtsCollection.XmlCtsCitation.ingest(
                self.response,
                xpath=".//ti:citation[not(ancestor::ti:citation)]"
            )



"""
.. module:: MyCapytain.resolvers.prototypes
   :synopsis: Resolver Prototype

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""


class Resolver(object):
    """ Resolver provide a native python API which returns python objects.

    Initiation of resolvers are dependent on the implementation of the prototype

    """
    def getMetadata(self, objectId=None, **filters):
        """ Request metadata about a text or a collection

        :param objectId: Object Identifier to filter on
        :type objectId: str
        :param filters: Kwargs parameters.
        :type filters: dict
        :return: Collection
        """
        raise NotImplementedError()

    def getTextualNode(self, textId, subreference=None, prevnext=False, metadata=False):
        """ Retrieve a text node from the API

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :param prevnext: Retrieve graph representing previous and next passage
        :type prevnext: boolean
        :param metadata: Retrieve metadata about the passage and the text
        :type metadata: boolean
        :return: CapitainsCtsPassage
        :rtype: CapitainsCtsPassage
        """
        raise NotImplementedError()

    def getSiblings(self, textId, subreference):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: Tuple of references
        :rtype: (str, str)
        """
        raise NotImplementedError()

    def getReffs(self, textId, level=1, subreference=None):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param level: Depth for retrieval
        :type level: int
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: List of references
        :rtype: [str]
        """
        raise NotImplementedError()



from setuptools import setup, find_packages

import MyCapytain

setup(
  name='MyCapytain',
  version=MyCapytain.__version__,
  description='Library for CTS APIs and CapiTainS guidelines in Python',
  url='http://github.com/Capitains/MyCapytain',
  author='Thibault Clerice',
  author_email='leponteineptique@gmail.com',
  license='Mozilla Public License Version 2.0',
  packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
  install_requires=[
    "requests>=2.8.1",
    "six>=1.10.0",
    "lxml>=3.6.4",
    "future>=0.16.0",
    "rdflib-jsonld>=0.4.0"
  ],
  tests_require=[
    "mock>=2.0.0",
    "responses>=0.8.1",
    "xmlunittest>=0.3.2"
  ],
  extras_require={
    "DOC": ["Sphinx==1.3.1"]
  },
  test_suite="tests",
  zip_safe=False,
  classifiers=[
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Information Technology",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Topic :: Software Development :: Libraries",
    "Topic :: Text Processing :: Markup :: XML",
    "Topic :: Text Processing :: General"
  ]
)



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.retrievers.protoypes
   :synopsis: Prototypes of APIs endpoint

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>


"""


class API(object):
    """
    API Prototype object

    :param self: Object
    :type self: API
    :param endpoint: URL of the API
    :type endpoint: text

    :ivar endpoint: Url of the endpoint
    """

    def __init__(self, endpoint):
        """ Instantiate an API class
        """
        self.endpoint = endpoint


class CitableTextServiceRetriever(API):
    """ Citable CtsTextMetadata Service retrievers should have at least have some of the following properties

    """

    def getMetadata(self, objectId=None, **filters):
        """ Request metadata about a text or a collection

        :param objectId: CtsTextMetadata Identifier
        :param filters: Kwargs parameters. URN and Inv are available
        :return: Metadata of text from an API or the likes as bytes
        """
        raise NotImplementedError

    def getTextualNode(self, textId, subreference=None, prevnext=False, metadata=False):
        """ Retrieve a text node from the API

        :param textId: CtsTextMetadata Identifier
        :param subreference: CapitainsCtsPassage Reference
        :param prevnext: Retrieve graph representing previous and next passage
        :param metadata: Retrieve metadata about the passage and the text
        :return: CtsTextMetadata of a CapitainsCtsPassage from an API or the likes as bytes
        """
        raise NotImplementedError

    def getSiblings(self, textId, subreference):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :param subreference: CapitainsCtsPassage Reference
        :return: Siblings references from an API or the likes as bytes
        """
        raise NotImplementedError

    def getReffs(self, textId, level=1, subreference=None):
        """ Retrieve the siblings of a textual node

        :param textId: CtsTextMetadata Identifier
        :type textId: str
        :param level: Depth for retrieval
        :type level: int
        :param subreference: CapitainsCtsPassage Reference
        :type subreference: str
        :return: List of references
        :rtype: [str]
        """
        raise NotImplementedError


class CtsRetriever(CitableTextServiceRetriever):
    """ CTS API Endpoint Prototype
    """
    def getCapabilities(self, inventory):
        """ Retrieve the inventory information of an API

        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getValidReff(self, urn, inventory, level=1):
        """ Retrieve valid urn-references for a text

        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param level: Depth of references expected
        :type level: int
        :rtype: str
        """
        raise NotImplementedError()

    def getFirstUrn(self, urn, inventory):
        """ Retrieve the first passage urn of a text

        :param urn: URN identifying the text
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getPrevNextUrn(self, urn, inventory):
        """ Retrieve the previous and next passage urn of one passage

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getLabel(self, urn, inventory):
        """ Retrieve informations about a CTS Urn

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :rtype: str
        """
        raise NotImplementedError()

    def getPassage(self, urn, inventory, context=None):
        """ Retrieve a passage

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn, \
         immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        raise NotImplementedError()

    def getPassagePlus(self, urn, inventory, context=None):
        """ Retrieve a passage and informations about it

        :param urn: URN identifying the text's passage (Minimum depth : 1)
        :type urn: text
        :param inventory: Name of the inventory
        :type inventory: text
        :param context: Number of citation units at the same level of the citation hierarchy as the requested urn,\
         immediately preceding and immediately following the requested urn to include in the reply
        :type context: int
        :rtype: str
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
.. module:: MyCapytain.resources.prototypes.metadata
   :synopsis: Definition of Metadata Type Objects

.. moduleauthor:: Thibault Clérice <leponteineptique@gmail.com>

"""

from MyCapytain.common.metadata import Metadata
from MyCapytain.errors import UnknownCollection
from MyCapytain.common.utils import Subgraph, LiteralToDict
from MyCapytain.common.constants import RDF_NAMESPACES, RDFLIB_MAPPING, Mimetypes, get_graph
from MyCapytain.common.base import Exportable
from rdflib import URIRef, RDF, Literal, Graph, RDFS
from rdflib.namespace import SKOS, DC


class Collection(Exportable):
    """ Collection represents any resource's metadata. It has members and parents

    :ivar properties: Properties of the collection
    :type properties: dict
    :ivar parents: Parent of the node from the direct parent to the highest ascendant
    :type parents: [Collection]
    :ivar metadata: Metadata
    :type metadata: Metadata
    """
    TYPE_URI = URIRef(RDF_NAMESPACES.DTS.collection)
    MODEL_URI = URIRef(RDF_NAMESPACES.DTS.collection)
    EXPORT_TO = [Mimetypes.JSON.LD, Mimetypes.JSON.DTS.Std, Mimetypes.XML.RDF]

    def __init__(self, identifier="", *args, **kwargs):
        super(Collection, self).__init__(identifier, *args, **kwargs)
        self.__graph__ = get_graph()

        self.__node__ = URIRef(identifier)
        self.__metadata__ = Metadata(node=self.asNode())
        self.__capabilities__ = Metadata.getOr(self.asNode(), RDF_NAMESPACES.DTS.capabilities)

        self.graph.set((self.asNode(), RDF.type, self.TYPE_URI))
        self.graph.set((self.asNode(), RDF_NAMESPACES.DTS.model, self.MODEL_URI))

        self.graph.addN(
            [
                (self.asNode(), RDF_NAMESPACES.DTS.capabilities, self.capabilities.asNode(), self.graph)
            ]
        )

        self.__parent__ = None
        self.__children__ = {}

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.id)

    @property
    def version(self):
        for x in self.graph.objects(self.asNode(), RDF_NAMESPACES.DTS.version):
            return x

    @version.setter
    def version(self, value):
        if not isinstance(value, Literal):
            value = Literal(value)
        self.graph.set((self.asNode(), RDF_NAMESPACES.DTS.version, value))

    @property
    def type(self):
        return list(self.graph.objects(self.asNode(), RDF.type))[0]

    @type.setter
    def type(self, value):
        if not isinstance(value, URIRef):
            value = URIRef(value)
        self.graph.set((self.asNode(), RDF.type, value))

    @property
    def model(self):
        return list(self.graph.objects(self.asNode(), RDF_NAMESPACES.DTS.model))[0]

    @model.setter
    def model(self, value):
        if not isinstance(value, URIRef):
            value = URIRef(value)
        self.graph.set((self.asNode(), RDF_NAMESPACES.DTS.model, value))

    @property
    def size(self):
        return len(self.members)

    # Graph Related Properties
    @property
    def graph(self):
        """ RDFLib Graph space

        :rtype: Graph
        """
        return self.__graph__

    @property
    def metadata(self):
        return self.__metadata__

    @property
    def capabilities(self):
        return self.__capabilities__

    def asNode(self):
        """ Node representation of the collection in the graph

        :rtype: URIRef
        """
        return self.__node__

    @property
    def id(self):
        return str(self.asNode())

    def get_label(self, lang=None):
        """ Return label for given lang or any default

        :param lang: Language to request
        :return: Label value
        :rtype: Literal
        """
        x = None
        if lang is None:
            for obj in self.graph.objects(self.asNode(), RDFS.label):
                return obj
        for obj in self.graph.objects(self.asNode(), RDFS.label):
            x = obj
            if x.language == lang:
                return x
        return x

    def set_label(self, label, lang):
        """ Add the label of the collection in given lang

        :param label: Label Value
        :param lang:  Language code
        """
        self.metadata.add(SKOS.prefLabel, Literal(label, lang=lang))
        self.graph.addN([
            (self.asNode(), RDFS.label, Literal(label, lang=lang), self.graph),
        ])

    @property
    def children(self):
        """ Dictionary of childrens {Identifier: Collection}

        :rtype: dict
        """
        return self.__children__

    @property
    def parents(self):
        """ Iterator to find parents of current collection, from closest to furthest

        :rtype: Generator[:class:`Collection`]
        """
        p = self.parent
        parents = []
        while p is not None:
            parents.append(p)
            p = p.parent
        return parents

    @property
    def parent(self):
        """ Parent of current object

        :rtype: Collection
        """
        return self.__parent__

    @parent.setter
    def parent(self, parent):
        """ Parents

        :param parent: Parent to set for the object
        :type parent: Collection
        :return:
        """
        self.__parent__ = parent
        self.graph.set(
            (self.asNode(), RDF_NAMESPACES.DTS.parent, parent.asNode())
        )
        parent.__add_member__(self)

    def __add_member__(self, member):
        """ Does not add member if it already knows it.

        .. warning:: It should not be called !

        :param member: Collection to add to members
        """
        if member.id in self.children:
            return None
        else:
            self.children[member.id] = member
            #self.graph.add((self.asNode(), RDF_NAMESPACES.DTS.child, member.asNode()))

    def __getitem__(self, key):
        """ Retrieve an item by its ID in the tree of a collection

        :param key: Key of the object to delete
        :return: Collection identified by the item
        """
        for obj in self.descendants + [self]:
            if obj.id == key:
                return obj
        raise UnknownCollection("%s is not part of this object" % key)

    def __delitem__(self, key):
        """ Delete a children of the collection

        :param key: Key of the object to delete
        """
        item = self[key]
        # Delete the graph Item
        self.graph.remove((item.asNode(), None, None))
        self.graph.remove((None, None, item.asNode()))
        self.metadata.remove()
        self.metadata.unlink()
        # Delete the Python item
        if len(item.parents) > 0:
            del item.parents[0].children[item.id]

    def __contains__(self, item):
        """ Retrieve an item by its ID in the tree of a collection

        :param item:
        :return: Collection identified by the item
        """
        for obj in self.descendants + [self]:
            if obj.id == item:
                return True
        return False

    @property
    def readable(self):
        """ Readable property should return elements where the element can be queried for getPassage / getReffs
        """
        return False

    @property
    def members(self):
        """ Children of the collection's item

        :rtype: [Collection]
        """
        return list(self.children.values())

    @property
    def descendants(self):
        """ Any descendant (no max level) of the collection's item

        :rtype: [Collection]
        """
        return self.members + \
            [submember for member in self.members for submember in member.descendants]

    @property
    def readableDescendants(self):
        """ List of element available which are readable

        :rtype: [Collection]
        """
        return [member for member in self.descendants if member.readable]

    def __namespaces_header__(self, cpt=None):
        """ Generates Namespaces Header given the graph

        :return: Dictionary with XMLNS prefix and uri as key and values
        """
        nm = self.graph.namespace_manager
        bindings = {}
        for predicate in set(self.graph.predicates()):
            prefix, namespace, name = nm.compute_qname(predicate)
            if prefix != "":
                bindings["xmlns:" + prefix] = str(URIRef(namespace))#[:-1]
            else:
                bindings["xmlns"] = str(URIRef(namespace))#[:-1]
        if cpt is True:
            bindings["xmlns:cpt"] = str(RDF_NAMESPACES.CAPITAINS)

        # Small hard coded fix for namespace that were not thought for RDF
        for k, v in bindings.items():
            if v in ["http://chs.harvard.edu/xmlns/cts/"]:
                bindings[k] = v[:-1]

        return bindings

    def __export__(self, output=None, domain=""):
        """ Export the collection item in the Mimetype required.

        ..note:: If current implementation does not have special mimetypes, reuses default_export method

        :param output: Mimetype to export to (Uses MyCapytain.common.utils.Mimetypes)
        :type output: str
        :param domain: Domain (Necessary sometime to express some IDs)
        :type domain: str
        :return: Object using a different representation
        """

        if output == Mimetypes.JSON.DTS.Std:
            nm = self.graph.namespace_manager
            bindings = {}
            for predicate in set(self.graph.predicates()):
                prefix, namespace, name = nm.compute_qname(predicate)
                bindings[prefix] = str(URIRef(namespace))

            RDFSLabel = self.graph.qname(RDFS.label)
            RDFType = self.graph.qname(RDF.type)
            store = Subgraph(get_graph().namespace_manager)
            store.graphiter(self.graph, self.metadata, ascendants=0, descendants=1)
            metadata = {}
            for _, predicate, obj in store.graph:
                k = self.graph.qname(predicate)
                if k in metadata:
                    if isinstance(metadata[k], list):
                        metadata[k].append(LiteralToDict(obj))
                    else:
                        metadata[k] = [metadata[k], LiteralToDict(obj)]
                else:
                    metadata[k] = LiteralToDict(obj)
            o = {
                "@context": bindings,
                "@graph": {
                    "@id": self.id,
                    RDFType: str(self.type),
                    RDFSLabel: LiteralToDict(self.get_label()) or self.id,
                    self.graph.qname(RDF_NAMESPACES.DTS.size): len(self.members),
                    self.graph.qname(RDF_NAMESPACES.DTS.metadata): metadata
                }
            }
            version = self.version
            if version is not None:
                o["@graph"]["version"] = str(version)
            if len(self.members):
                o["@graph"][self.graph.qname(RDF_NAMESPACES.DTS.members)] = [
                    {
                        "@id": member.id,
                        RDFSLabel: LiteralToDict(member.get_label()) or member.id,
                        self.graph.qname(RDF_NAMESPACES.DTS.url): domain + member.id
                    }
                    for member in self.members
                ]
            if self.parent:
                o["@graph"][self.graph.qname(RDF_NAMESPACES.DTS.parents)] = [
                    {
                        "@id": member.id,
                        RDFSLabel: LiteralToDict(member.get_label()) or member.id,
                        self.graph.qname(RDF_NAMESPACES.DTS.url): domain + member.id
                    }
                    for member in self.parents
                ]
            del store
            return o
        elif output == Mimetypes.JSON.LD\
                or output == Mimetypes.XML.RDF:

            # We create a temp graph
            store = Subgraph(get_graph().namespace_manager)
            store.graphiter(self.graph, self.asNode(), ascendants=1, descendants=-1)

            o = store.serialize(format=RDFLIB_MAPPING[output], auto_compact=True, indent="")
            del store
            return o


class ResourceCollection(Collection):
    @property
    def lang(self):
        """ Languages this text is in

        :return: List of available languages
        """
        return str(self.graph.value(self.asNode(), DC.language))

    @lang.setter
    def lang(self, lang):
        """ Language this text is available in

        :param lang: Language to add
        :type lang: str
        """
        self.graph.set((self.asNode(), DC.language, Literal(lang)))

    def get_creator(self, lang=None):
        """ Get the DC Creator literal value

        :param lang: Language to retrieve
        :return: Creator string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.creator, lang=lang)

    def get_title(self, lang=None):
        """ Get the title of the object

        :param lang: Lang to retrieve
        :return: Title string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.title, lang=lang)

    def get_description(self, lang=None):
        """ Get the description of the object

        :param lang: Lang to retrieve
        :return: Description string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.description, lang=lang)

    def get_subject(self, lang=None):
        """ Get the subject of the object

        :param lang: Lang to retrieve
        :return: Subject string representation
        :rtype: Literal
        """
        return self.metadata.get_single(key=DC.subject, lang=lang)
